""" Contains different test cases for tests for reading TDMS files
"""
import numpy as np
import pytest

from nptdms.test.util import (
    GeneratedFile,
    hexlify_value,
    string_hexlify,
    segment_objects_metadata,
    channel_metadata,
    channel_metadata_with_no_data,
    channel_metadata_with_repeated_structure,
    root_metadata,
    group_metadata)


TDS_TYPE_INT8 = 1
TDS_TYPE_INT16 = 2
TDS_TYPE_INT32 = 3
TDS_TYPE_BOOL = 0x21
TDS_TYPE_COMPLEX64 = 0x08000c
TDS_TYPE_COMPLEX128 = 0x10000d
TDS_TYPE_FLOAT32 = 9
TDS_TYPE_FLOAT64 = 10
TDS_TYPE_FLOAT32_WITH_UNIT = 0x19
TDS_TYPE_FLOAT64_WITH_UNIT = 0x1A


_scenarios = []


def scenario(func):
    def as_param():
        result = func()
        return pytest.param(*result, id=func.__name__)
    _scenarios.append(as_param)
    return as_param


def get_scenarios():
    return [f() for f in _scenarios]


@scenario
def single_segment_with_one_channel():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 4),
        ),
        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def single_segment_with_two_channels():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def single_segment_with_two_channels_interleaved():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 3], dtype=np.int32),
        ('group', 'channel2'): np.array([2, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def single_segment_with_interleaved_data_of_different_width():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT8, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT16, 2),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
        ),
        "01" "01 00" "01 00 00 00"
        "02" "02 00" "02 00 00 00"
        "03" "03 00" "03 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 3], dtype=np.int8),
        ('group', 'channel2'): np.array([1, 2, 3], dtype=np.int16),
        ('group', 'channel3'): np.array([1, 2, 3], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def no_metadata_segment():
    """ Add a segment with two channels, then a second
        segment with the same metadata as before,
        so there is only the lead in and binary data
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocRawData", ),
        "",
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def object_with_no_metadata_in_segment():
    """ Re-use an object without setting any new metadata and
        re-using the data structure
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
            channel_metadata_with_repeated_structure("/'group'/'channel2'"),
        ),
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def add_new_channel():
    """ Add a new voltage channel, with the other two channels
        remaining unchanged, so only the new channel is in metadata section
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
        ),
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
        "09 00 00 00" "0A 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8], dtype=np.int32),
        ('group', 'channel3'): np.array([9, 10], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def repeated_objects_without_data_in_segment_and_added_object():
    """ Repeated objects with no data in new segment as well as a new channel with data
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
            channel_metadata_with_no_data("/'group'/'channel2'"),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
        ),
        "05 00 00 00" "06 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4], dtype=np.int32),
        ('group', 'channel3'): np.array([5, 6], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def increase_channel_size():
    """ In the second segment, increase the channel size of one channel
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 4),
        ),
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
        "09 00 00 00" "0A 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8, 9, 10], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def remove_a_channel():
    """ In the second segment, remove a channel.
        We need to write a new object list in this case
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
        ),
        "05 00 00 00" "06 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def alternating_data_objects_with_new_obj_list():
    """ Alternating segments with different objects with data
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata_with_no_data("/'group'/'channel2'"),
        ),
        "01 00 00 00" "02 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
            channel_metadata_with_no_data("/'group'/'channel2'"),
        ),
        "05 00 00 00" "06 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
            channel_metadata_with_repeated_structure("/'group'/'channel2'"),
        ),
        "07 00 00 00" "08 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def alternating_data_objects_reusing_obj_list():
    """ Alternating segments with different objects with data,
        reusing the object list from the last segment.
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata_with_no_data("/'group'/'channel2'"),
        ),
        "01 00 00 00" "02 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
            channel_metadata_with_no_data("/'group'/'channel2'"),
        ),
        "05 00 00 00" "06 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
            channel_metadata_with_repeated_structure("/'group'/'channel2'"),
        ),
        "07 00 00 00" "08 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5, 6], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 7, 8], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def chunked_segment():
    """ Add segment and then a repeated segment without
        any lead in or metadata, so data is read in chunks
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "00 00 00 00" "01 00 00 00"
        "0A 00 00 00" "0B 00 00 00"
        "02 00 00 00" "03 00 00 00"
        "0C 00 00 00" "0D 00 00 00"
    )
    test_file.add_segment(
        ("kTocRawData", ),
        "",
        "04 00 00 00" "05 00 00 00"
        "0E 00 00 00" "0F 00 00 00"
        "06 00 00 00" "07 00 00 00"
        "10 00 00 00" "11 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([0, 1, 2, 3, 4, 5, 6, 7], dtype=np.int32),
        ('group', 'channel2'): np.array([10, 11, 12, 13, 14, 15, 16, 17], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def chunked_interleaved_segment():
    """ Add interleaved segment and then a repeated segment without
        any lead in or metadata, so data is read in chunks
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
    )
    test_file.add_segment(
        ("kTocRawData", "kTocInterleavedData"),
        "",
        "07 00 00 00" "08 00 00 00"
        "05 00 00 00" "06 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "01 00 00 00" "02 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 3, 5, 7, 7, 5, 3, 1], dtype=np.int32),
        ('group', 'channel2'): np.array([2, 4, 6, 8, 8, 6, 4, 2], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def less_data_than_expected():
    """ Add segment and then a repeated segment without
        any lead in or metadata, so data is read in chunks,
        but the extra chunk does not have as much data as expected.
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "05 00 00 00" "06 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 5], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 6], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def less_data_than_expected_interleaved():
    """ Add interleaved data segment and then a repeated segment without
        any lead in or metadata, so data is read in chunks,
        but the extra chunk does not have as much data as expected.
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 3),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 3),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
        "09 00 00 00" "0A 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 3, 5, 7, 9], dtype=np.int32),
        ('group', 'channel2'): np.array([2, 4, 6, 8, 10], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def segment_with_zero_data_length():
    """ kTocRawData is set but data length is zero
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 0),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 0),
        ),
        ""
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def incomplete_last_segment():
    """ Test incomplete last segment, eg. if LabView crashed
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            root_metadata(),
            group_metadata(),
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
        "09 00 00 00" "0A 00 00 00"
        "0B 00 00 00" "0C 00 00 00"
    )
    test_file.add_segment(
        ("kTocRawData", ),
        "",
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00",
        incomplete=True
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 7, 8, 1, 2], dtype=np.int32),
        ('group', 'channel2'): np.array([3, 4, 9, 10, 3], dtype=np.int32),
        ('group', 'channel3'): np.array([5, 6, 11, 12], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def incomplete_last_row_of_interleaved_data():
    """ Test incomplete last row of interleaved data
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            root_metadata(),
            group_metadata(),
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
        "05 00 00 00" "06 00 00 00"
        "07 00 00 00" "08 00 00 00"
        "09 00 00 00" "0A 00 00 00"
        "0B 00 00 00",
        incomplete=True
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 3, 5, 7, 9], dtype=np.int32),
        ('group', 'channel2'): np.array([2, 4, 6, 8, 10], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def bool_data():
    """ Test reading a file with boolean valued data
    """
    expected_channel_data = np.array([False, True, False, True], dtype=np.dtype(np.bool_))

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'bool_channel'", TDS_TYPE_BOOL, 2),
        ),
        "00 01 00 01"
    )
    expected_data = {
        ('group', 'bool_channel'): expected_channel_data,
    }
    return test_file, expected_data


@scenario
def float_data():
    """ Test reading a file with float valued data
    """
    single_arr = np.array([0.123, 0.234, 0.345, 0.456], dtype=np.float32)
    double_arr = np.array([0.987, 0.876, 0.765, 0.654], dtype=np.double)
    data = ""
    for num in single_arr[0:2]:
        data += hexlify_value("<f", num)
    for num in double_arr[0:2]:
        data += hexlify_value("<d", num)
    for num in single_arr[2:4]:
        data += hexlify_value("<f", num)
    for num in double_arr[2:4]:
        data += hexlify_value("<d", num)

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'single_channel'", TDS_TYPE_FLOAT32, 2),
            channel_metadata("/'group'/'double_channel'", TDS_TYPE_FLOAT64, 2),
        ),
        data
    )
    expected_data = {
        ('group', 'single_channel'): single_arr,
        ('group', 'double_channel'): double_arr,
    }
    return test_file, expected_data


@scenario
def float_data_with_unit():
    """ Test reading a file with float valued data with units

        These are the same as normal floating point data but have a 'unit_string' property
    """
    single_arr = np.array([0.123, 0.234, 0.345, 0.456], dtype=np.float32)
    double_arr = np.array([0.987, 0.876, 0.765, 0.654], dtype=np.double)
    data = ""
    for num in single_arr[0:2]:
        data += hexlify_value("<f", num)
    for num in double_arr[0:2]:
        data += hexlify_value("<d", num)
    for num in single_arr[2:4]:
        data += hexlify_value("<f", num)
    for num in double_arr[2:4]:
        data += hexlify_value("<d", num)

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'single_channel'", TDS_TYPE_FLOAT32_WITH_UNIT, 2),
            channel_metadata("/'group'/'double_channel'", TDS_TYPE_FLOAT64_WITH_UNIT, 2),
        ),
        data
    )
    expected_data = {
        ('group', 'single_channel'): single_arr,
        ('group', 'double_channel'): double_arr,
    }
    return test_file, expected_data


@scenario
def complex_data():
    """ Test reading a file with complex valued data
    """
    complex_single_arr = np.array([1+2j, 3+4j], dtype=np.complex64)
    complex_double_arr = np.array([5+6j, 7+8j], dtype=np.complex128)
    data = ""
    for num in complex_single_arr:
        data += hexlify_value("<f", num.real)
        data += hexlify_value("<f", num.imag)
    for num in complex_double_arr:
        data += hexlify_value("<d", num.real)
        data += hexlify_value("<d", num.imag)

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'complex_single_channel'", TDS_TYPE_COMPLEX64, 2),
            channel_metadata("/'group'/'complex_double_channel'", TDS_TYPE_COMPLEX128, 2),
        ),
        data
    )
    expected_data = {
        ('group', 'complex_single_channel'): complex_single_arr,
        ('group', 'complex_double_channel'): complex_double_arr,
    }
    return test_file, expected_data


@scenario
def scaled_data():
    properties = {
        "NI_Number_Of_Scales":
            (3, "01 00 00 00"),
        "NI_Scale[0]_Scale_Type":
            (0x20, hexlify_value("<I", len("Linear")) + string_hexlify("Linear")),
        "NI_Scale[0]_Linear_Slope":
            (10, hexlify_value("<d", 2.0)),
        "NI_Scale[0]_Linear_Y_Intercept":
            (10, hexlify_value("<d", 10.0))
    }
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2, properties),
        ),
        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([12, 14, 16, 18], dtype=np.float64),
    }
    return test_file, expected_data


@scenario
def timestamp_data():
    """Test reading contiguous timestamp data
    """

    times = [
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
    ]

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'Group'/'TimeChannel1'", 0x44, 2),
            channel_metadata("/'Group'/'TimeChannel2'", 0x44, 2),
        ),
        timestamp_data_chunk(times)
    )

    expected_data = {
        ('Group', 'TimeChannel1'): np.array([times[0], times[1]]),
        ('Group', 'TimeChannel2'): np.array([times[2], times[3]]),
    }

    return test_file, expected_data


@scenario
def interleaved_timestamp_data():
    """Test reading interleaved timestamp data
    """

    times = [
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
    ]

    metadata = (
        # Number of objects
        "02 00 00 00"
        # Length of the object path
        "17 00 00 00")
    metadata += string_hexlify("/'Group'/'TimeChannel1'")
    metadata += (
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "44 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")
    metadata += (
        "17 00 00 00")
    metadata += string_hexlify("/'Group'/'TimeChannel2'")
    metadata += (
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "44 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
    test_file.add_segment(toc, metadata, timestamp_data_chunk(times))

    expected_data = {
        ('Group', 'TimeChannel1'): np.array([times[0], times[2]]),
        ('Group', 'TimeChannel2'): np.array([times[1], times[3]]),
    }

    return test_file, expected_data


@scenario
def interleaved_timestamp_and_numpy_data():
    """Test reading timestamp data interleaved with a standard numpy data type
    """

    times = [
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
    ]

    metadata = (
        # Number of objects
        "02 00 00 00"
        # Length of the object path
        "16 00 00 00")
    metadata += string_hexlify("/'Group'/'TimeChannel'")
    metadata += (
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "44 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")
    metadata += (
        "15 00 00 00")
    metadata += string_hexlify("/'Group'/'IntChannel'")
    metadata += (
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")

    data = (timestamp_data_chunk([times[0]]) +
            "01 00 00 00" +
            timestamp_data_chunk([times[1]]) +
            "02 00 00 00")

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
    test_file.add_segment(toc, metadata, data)

    expected_data = {
        ('Group', 'TimeChannel'): np.array([times[0], times[1]]),
        ('Group', 'IntChannel'): np.array([1, 2], dtype=np.dtype('int32')),
    }

    return test_file, expected_data


@scenario
def segment_without_data():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
        ),
        ""
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
        ),
        "01 00 00 00" "02 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
        ),
        ""
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
        ),
        "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
    }
    return test_file, expected_data


@scenario
def channel_without_data():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
        ),
        ""
    )
    expected_data = {
        ('group', 'channel1'): np.array([], dtype=np.dtype('int32')),
    }
    return test_file, expected_data


@scenario
def channel_without_data_or_data_type():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata_with_no_data("/'group'/'channel1'"),
        ),
        ""
    )
    expected_data = {
        ('group', 'channel1'): np.array([], dtype=np.dtype('void')),
    }
    return test_file, expected_data


@scenario
def extra_padding_after_metadata():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
        ) + "00 00 00 00 00 00 00 00",
        "01 00 00 00" "02 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
        ) + "00 00 00 00 00 00 00 00",
        "03 00 00 00" "04 00 00 00"
    )
    expected_data = {
        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
    }
    return test_file, expected_data


def timestamp_data_chunk(times):
    return "".join(timestamp_hexlify(time) for time in times)


def timestamp_hexlify(time):
    epoch = np.datetime64('1904-01-01T00:00:00')

    relative_time = time - epoch
    total_seconds = int(relative_time / np.timedelta64(1, 's'))

    remainder = relative_time - np.timedelta64(total_seconds, 's')
    microseconds = int(remainder / np.timedelta64(1, 'us'))
    fraction = int(float(microseconds) * 2 ** 58 / 5 ** 6)

    return hexlify_value("<Q", fraction) + hexlify_value("<q", total_seconds)



"""Test exporting TDMS data to Pandas"""

from datetime import datetime
import numpy as np
import pytest
try:
    import pandas
except ImportError:
    pytest.skip("Skipping Pandas tests as Pandas is not installed", allow_module_level=True)

from nptdms import TdmsFile
from nptdms.test import scenarios
from nptdms.test.test_daqmx import daqmx_channel_metadata, daqmx_scaler_metadata
from nptdms.test.util import (
    GeneratedFile,
    basic_segment,
    channel_metadata,
    channel_metadata_with_no_data,
    string_hexlify,
    segment_objects_metadata,
    hexlify_value
)


def assert_within_tol(a, b, tol=1.0e-10):
    assert abs(a - b) < tol


def timed_segment():
    """TDMS segment with one group and two channels,
    each with time properties"""

    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "03 00 00 00"
        # Length of the first object path
        "08 00 00 00"
        # Object path (/'Group')
        "2F 27 47 72"
        "6F 75 70 27"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "02 00 00 00"
        # Name length
        "04 00 00 00"
        # Property name (prop)
        "70 72 6F 70"
        # Property data type (string)
        "20 00 00 00"
        # Length of string value
        "05 00 00 00"
        # Value
        "76 61 6C 75 65"
        # Length of second property name
        "03 00 00 00"
        # Property name (num)
        "6E 75 6D"
        # Data type of property
        "03 00 00 00"
        # Value
        "0A 00 00 00"
        # Length of the second object path
        "13 00 00 00"
        # Second object path (/'Group'/'Channel1')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 31 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties
        "03 00 00 00"
        # Set time properties for the first channel
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 2.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1) +
        "0D 00 00 00" +
        string_hexlify('wf_start_time') +
        "44 00 00 00" +
        hexlify_value("<Q", 0) +
        hexlify_value("<q", 3524551547) +
        # Length of the third object path
        "13 00 00 00"
        # Third object path (/'Group'/'Channel2')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 32 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties
        "03 00 00 00"
        # Set time properties for the second channel
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 2.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1) +
        "0D 00 00 00" +
        string_hexlify('wf_start_time') +
        "44 00 00 00" +
        hexlify_value("<Q", 0) +
        hexlify_value("<q", 3524551547))
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "03 00 00 00"
        "04 00 00 00"
    )
    return toc, metadata, data


@pytest.mark.parametrize('lazy_load', [True, False])
def test_file_as_dataframe(lazy_load):
    """Test converting file to Pandas dataframe"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())

    df = on_test_file(test_file, lazy_load, lambda tdms_data: tdms_data.as_dataframe())

    assert len(df) == 2
    assert "/'Group'/'Channel1'" in df.keys()
    assert "/'Group'/'Channel2'" in df.keys()

    assert (df["/'Group'/'Channel1'"] == [1, 2]).all()


def test_file_as_dataframe_without_time():
    """Converting file to dataframe with time index should raise when
    time properties aren't present"""

    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    with pytest.raises(KeyError):
        tdms_data.as_dataframe(time_index=True)


def test_file_as_dataframe_with_time():
    """Test converting file to Pandas dataframe with a time index"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())
    tdms_data = test_file.load()

    df = tdms_data.as_dataframe(time_index=True)

    assert len(df.index) == 2
    assert_within_tol(df.index[0], 2.0)
    assert_within_tol(df.index[1], 2.1)


def test_file_as_dataframe_with_absolute_time():
    """Convert file to Pandas dataframe with absolute time index"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())
    tdms_data = test_file.load()

    df = tdms_data.as_dataframe(time_index=True, absolute_time=True)

    expected_start = datetime(2015, 9, 8, 10, 5, 49)
    assert (df.index == expected_start)[0]


@pytest.mark.parametrize('lazy_load', [True, False])
def test_group_as_dataframe(lazy_load):
    """Convert a group to dataframe"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())

    df = on_test_file(test_file, lazy_load, lambda tdms_data: tdms_data["Group"].as_dataframe())

    assert len(df) == 2
    assert len(df.keys()) == 2
    assert "Channel1" in df.keys()
    assert "Channel2" in df.keys()
    assert (df["Channel1"] == [1, 2]).all()
    assert (df["Channel2"] == [3, 4]).all()


@pytest.mark.parametrize('lazy_load', [True, False])
def test_channel_as_dataframe(lazy_load):
    """Convert a channel to dataframe"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())

    df = on_test_file(
        test_file, lazy_load, lambda tdms_data: tdms_data["Group"]["Channel2"].as_dataframe())

    assert len(df) == 2
    assert len(df.keys()) == 1
    assert "/'Group'/'Channel2'" in df.keys()
    assert (df["/'Group'/'Channel2'"] == [3, 4]).all()


def test_channel_as_dataframe_with_time():
    """Convert a channel to dataframe with a time index"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())
    tdms_data = test_file.load()

    df = tdms_data["Group"]["Channel2"].as_dataframe(time_index=True)

    assert len(df.index) == 2
    assert_within_tol(df.index[0], 2.0)
    assert_within_tol(df.index[1], 2.1)


def test_channel_as_dataframe_without_time():
    """Converting channel to dataframe should work correctly"""

    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    df = tdms_data["Group"]["Channel2"].as_dataframe()

    assert len(df.index) == 2
    assert len(df.values) == 2
    assert_within_tol(df.index[0], 0)
    assert_within_tol(df.index[1], 1)
    assert_within_tol(df.values[0], 3.0)
    assert_within_tol(df.values[1], 4.0)


def test_channel_as_dataframe_with_absolute_time():
    """Convert channel to Pandas dataframe with absolute time index"""

    test_file = GeneratedFile()
    test_file.add_segment(*timed_segment())
    tdms_data = test_file.load()

    df = tdms_data["Group"]["Channel1"].as_dataframe(time_index=True, absolute_time=True)

    expected_start = datetime(2015, 9, 8, 10, 5, 49)
    assert (df.index == expected_start)[0]


@pytest.mark.parametrize('lazy_load', [True, False])
def test_channel_as_dataframe_with_raw_data(lazy_load):
    """Convert channel to Pandas dataframe with absolute time index"""

    test_file, _ = scenarios.scaled_data().values
    expected_raw_data = np.array([1, 2, 3, 4], dtype=np.int32)

    df = on_test_file(
        test_file, lazy_load, lambda tdms_data: tdms_data["group"]["channel1"].as_dataframe(scaled_data=False))

    np.testing.assert_equal(df["/'group'/'channel1'"], expected_raw_data)


@pytest.mark.parametrize('lazy_load', [True, False])
def test_raw_daqmx_channel_export(lazy_load):
    """ Test exporting raw daqmx data for a channel
    """

    scaler_metadata = [
        daqmx_scaler_metadata(0, 3, 0),
        daqmx_scaler_metadata(1, 3, 2)]
    metadata = segment_objects_metadata(
        daqmx_channel_metadata("Channel1", 4, [4], scaler_metadata))
    data = (
        # Data for segment
        "01 00"
        "11 00"
        "02 00"
        "12 00"
        "03 00"
        "13 00"
        "04 00"
        "14 00"
    )

    test_file = GeneratedFile()
    segment_toc = (
        "kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocDAQmxRawData")
    test_file.add_segment(segment_toc, metadata, data)
    dataframe = on_test_file(
        test_file, lazy_load, lambda tdms_data: tdms_data["Group"]["Channel1"].as_dataframe(scaled_data=False))

    expected_data = {
        0: np.array([1, 2, 3, 4], dtype=np.int16),
        1: np.array([17, 18, 19, 20], dtype=np.int16),
    }
    assert dataframe["/'Group'/'Channel1'[0]"].dtype == np.int16
    assert dataframe["/'Group'/'Channel1'[1]"].dtype == np.int16
    np.testing.assert_equal(dataframe["/'Group'/'Channel1'[0]"], expected_data[0])
    np.testing.assert_equal(dataframe["/'Group'/'Channel1'[1]"], expected_data[1])


def test_export_with_empty_channels():
    """Convert a group to dataframe when a channel has empty data and void data type"""

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2),
            channel_metadata_with_no_data("/'group'/'channel2'"),
        ),
        "01 00 00 00 02 00 00 00"
    )

    tdms_data = test_file.load()

    df = tdms_data["group"].as_dataframe()
    assert len(df) == 2
    assert len(df.keys()) == 2
    assert "channel1" in df.keys()
    assert "channel2" in df.keys()
    assert (df["channel1"] == [1, 2]).all()
    assert len(df["channel2"]) == 2
    assert np.isnan(df["channel2"]).all()


def on_test_file(test_file, lazy_load, func):
    if lazy_load:
        with test_file.get_tempfile() as temp_file:
            with TdmsFile.open(temp_file) as tdms_file:
                return func(tdms_file)
    else:
        return func(test_file.load())



import logging


class LogManager(object):
    """ Manages multiple loggers from different modules
    """

    def __init__(self):
        self.log_level = logging.WARNING

        self.console_handler = logging.StreamHandler()
        self.console_handler.setLevel(self.log_level)

        self.formatter = logging.Formatter(
            '[%(name)s %(levelname)s] %(message)s')
        self.console_handler.setFormatter(self.formatter)

        self.loggers = {}

    def get_logger(self, module_name):
        """ Return a logger for a module
        """
        log = logging.getLogger(module_name)
        log.setLevel(self.log_level)
        log.addHandler(self.console_handler)
        self.loggers[module_name] = log
        return log

    def set_level(self, level):
        """ Set the log level for all loggers that have been created
        """
        self.log_level = level
        self.console_handler.setLevel(level)
        for log in self.loggers.values():
            log.setLevel(level)


log_manager = LogManager()



from collections import OrderedDict
import numpy as np


def from_tdms_file(tdms_file, time_index=False, absolute_time=False, scaled_data=True):
    """
    Converts the TDMS file to a DataFrame. DataFrame columns are named using the TDMS object paths.

    :param tdms_file: TDMS file object to convert.
    :param time_index: Whether to include a time index for the dataframe.
    :param absolute_time: If time_index is true, whether the time index
        values are absolute times or relative to the start time.
    :param scaled_data: By default the scaled data will be used.
        Set to False to use raw unscaled data.
    :return: The full TDMS file data.
    :rtype: pandas.DataFrame
    """

    channels_to_export = OrderedDict()
    for group in tdms_file.groups():
        for channel in group.channels():
            channels_to_export[channel.path] = channel
    return _channels_to_dataframe(channels_to_export, time_index, absolute_time, scaled_data)


def from_group(group, time_index=False, absolute_time=False, scaled_data=True):
    """
    Converts a TDMS group object to a DataFrame. DataFrame columns are named using the channel names.

    :param group: Group object to convert.
    :param time_index: Whether to include a time index for the dataframe.
    :param absolute_time: If time_index is true, whether the time index
        values are absolute times or relative to the start time.
    :param scaled_data: By default the scaled data will be used.
        Set to False to use raw unscaled data.
    :return: The TDMS object data.
    :rtype: pandas.DataFrame
    """

    channels_to_export = OrderedDict((ch.name, ch) for ch in group.channels())
    return _channels_to_dataframe(channels_to_export, time_index, absolute_time, scaled_data)


def from_channel(channel, time_index=False, absolute_time=False, scaled_data=True):
    """
    Converts the TDMS channel to a DataFrame

    :param channel: Channel object to convert.
    :param time_index: Whether to include a time index for the dataframe.
    :param absolute_time: If time_index is true, whether the time index
        values are absolute times or relative to the start time.
    :param scaled_data: By default the scaled data will be used.
        Set to False to use raw unscaled data.
    :return: The TDMS object data.
    :rtype: pandas.DataFrame
    """

    channels_to_export = {channel.path: channel}
    return _channels_to_dataframe(channels_to_export, time_index, absolute_time, scaled_data)


def _channels_to_dataframe(channels_to_export, time_index=False, absolute_time=False, scaled_data=True):
    import pandas as pd

    dataframe_dict = OrderedDict()
    for column_name, channel in channels_to_export.items():
        index = channel.time_track(absolute_time) if time_index else None
        if scaled_data:
            dataframe_dict[column_name] = pd.Series(data=_array_for_pd(channel[:]), index=index)
        elif channel.scaler_data_types:
            # Channel has DAQmx raw data
            raw_data = channel.read_data(scaled=False)
            for scale_id, scaler_data in raw_data.items():
                scaler_column_name = column_name + "[{0:d}]".format(scale_id)
                dataframe_dict[scaler_column_name] = pd.Series(data=scaler_data, index=index)
        else:
            # Raw data for normal TDMS file
            raw_data = channel.read_data(scaled=False)
            dataframe_dict[column_name] = pd.Series(data=_array_for_pd(raw_data), index=index)
    return pd.DataFrame.from_dict(dataframe_dict)


def _array_for_pd(array):
    """ Convert data array to a format suitable for a Pandas dataframe
    """
    if np.issubdtype(array.dtype, np.dtype('void')):
        # If dtype is void then the array must also be empty.
        # Pandas doesn't like void data types, so these are converted to empty float64 arrays
        # and Pandas will fill values with NaN
        return np.empty(0, dtype='float64')
    return array



""" Utilities for testing TDMS reading
"""

import binascii
from contextlib import contextmanager
import os
from io import BytesIO
import struct
import tempfile
import numpy as np

from nptdms import tdms


def string_hexlify(input_string):
    """Return hex string representation of string"""
    return binascii.hexlify(input_string.encode('utf-8')).decode('utf-8')


def hexlify_value(struct_type, value):
    """Return hex string representation of a value"""
    return binascii.hexlify(struct.pack(struct_type, value)).decode('utf-8')


def segment_objects_metadata(*args):
    """ Metadata for multiple objects in a segment
    """
    num_objects_hex = hexlify_value("<I", len(args))
    return num_objects_hex + "".join(args)


def root_metadata():
    return (
        # Length of the object path
        "01 00 00 00"
        # Object path (/)
        "2F"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "00 00 00 00")


def group_metadata():
    return (
        # Length of the object path
        "08 00 00 00"
        # Object path (/'Group')
        "2F 27 47 72"
        "6F 75 70 27"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "00 00 00 00")


def channel_metadata(channel_name, data_type, num_values, properties=None):
    return (
        # Length of the object path
        hexlify_value('<I', len(channel_name)) +
        # Object path
        string_hexlify(channel_name) +
        # Length of index information
        "14 00 00 00" +
        # Raw data data type
        hexlify_value('<I', data_type) +
        # Dimension
        "01 00 00 00" +
        # Number of raw data values
        hexlify_value('<Q', num_values) +
        hex_properties(properties)
    )


def hex_properties(properties):
    if properties is None:
        properties = {}
    props_hex = hexlify_value('<I', len(properties))
    for (prop_name, (prop_type, prop_value)) in properties.items():
        props_hex += hexlify_value('<I', len(prop_name))
        props_hex += string_hexlify(prop_name)
        props_hex += hexlify_value('<I', prop_type)
        props_hex += prop_value
    return props_hex


def channel_metadata_with_repeated_structure(channel_name):
    return (
        # Length of the object path
        hexlify_value('<I', len(channel_name)) +
        # Object path
        string_hexlify(channel_name) +
        # Raw data index header meaning repeat previous data structure
        "00 00 00 00" +
        # Number of properties (0)
        "00 00 00 00"
    )


def channel_metadata_with_no_data(channel_name):
    return (
        # Length of the object path
        hexlify_value('<I', len(channel_name)) +
        # Object path
        string_hexlify(channel_name) +
        # Raw data index header meaning no data in this segment
        "FF FF FF FF" +
        # Number of properties (0)
        "00 00 00 00"
    )


def basic_segment():
    """Basic TDMS segment with one group and two channels"""

    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "04 00 00 00"
        # Length of the first object path
        "08 00 00 00"
        # Object path (/'Group')
        "2F 27 47 72"
        "6F 75 70 27"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "02 00 00 00"
        # Name length
        "04 00 00 00"
        # Property name (prop)
        "70 72 6F 70"
        # Property data type (string)
        "20 00 00 00"
        # Length of string value
        "05 00 00 00"
        # Value
        "76 61 6C 75 65"
        # Length of second property name
        "03 00 00 00"
        # Property name (num)
        "6E 75 6D"
        # Data type of property
        "03 00 00 00"
        # Value
        "0A 00 00 00"
        # Length of the second object path
        "13 00 00 00"
        # Second object path (/'Group'/'Channel1')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 31 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00"
        # Length of the third object path
        "13 00 00 00"
        # Third object path (/'Group'/'Channel2')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 32 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of data values
        "02 00 00 00"
        "00 00 00 00"
        # Set time properties for the second channel
        "02 00 00 00"
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1) +
        # Length of the object path
        "01 00 00 00"
        # Object path (/)
        "2F"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "01 00 00 00"
        # Length of property name
        "03 00 00 00"
        # Property name (num)
        "6E 75 6D"
        # Data type of property
        "03 00 00 00"
        # Value
        "0F 00 00 00"
    )
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "03 00 00 00"
        "04 00 00 00"
    )
    return toc, metadata, data


class GeneratedFile(object):
    """Generate a TDMS file for testing"""

    def __init__(self):
        self._content = []

    def add_segment(self, toc, metadata, data, incomplete=False, binary_data=False, version=4713):
        metadata_bytes = _hex_to_bytes(metadata)
        data_bytes = data if binary_data else _hex_to_bytes(data)
        if toc is not None:
            lead_in = b'TDSm'
            toc_mask = 0
            for toc_item in toc:
                if toc_item == "kTocMetaData":
                    toc_mask = toc_mask | 1 << 1
                elif toc_item == "kTocRawData":
                    toc_mask = toc_mask | 1 << 3
                elif toc_item == "kTocDAQmxRawData":
                    toc_mask = toc_mask | 1 << 7
                elif toc_item == "kTocInterleavedData":
                    toc_mask = toc_mask | 1 << 5
                elif toc_item == "kTocBigEndian":
                    toc_mask = toc_mask | 1 << 6
                elif toc_item == "kTocNewObjList":
                    toc_mask = toc_mask | 1 << 2
                else:
                    raise ValueError("Unrecognised TOC value: %s" % toc_item)
            lead_in += struct.pack('<i', toc_mask)
            lead_in += struct.pack('<l', version)
            next_segment_offset = len(metadata_bytes) + len(data_bytes)
            raw_data_offset = len(metadata_bytes)
            if incomplete:
                lead_in += _hex_to_bytes('FF' * 8)
            else:
                lead_in += struct.pack('<Q', next_segment_offset)
            lead_in += struct.pack('<Q', raw_data_offset)
        else:
            lead_in = b''
        self._content.append((lead_in, metadata_bytes, data_bytes))

    def get_tempfile(self, **kwargs):
        named_file = tempfile.NamedTemporaryFile(suffix=".tdms", **kwargs)
        file = named_file.file
        file.write(self._get_contents())
        file.seek(0)
        return named_file

    @contextmanager
    def get_tempfile_with_index(self):
        directory = tempfile.mkdtemp()
        tdms_path = os.path.join(directory, 'test_file.tdms')
        tdms_index_path = os.path.join(directory, 'test_file.tdms_index')
        with open(tdms_path, 'wb') as file:
            file.write(self._get_contents())
        with open(tdms_index_path, 'wb') as file:
            file.write(self._get_index_contents())
        try:
            yield tdms_path
        finally:
            os.unlink(tdms_path)
            os.unlink(tdms_index_path)
            os.rmdir(directory)

    def load(self, *args, **kwargs):
        with tempfile.NamedTemporaryFile(suffix=".tdms") as named_file:
            file = named_file.file
            file.write(self._get_contents())
            file.seek(0)
            return tdms.TdmsFile(file, *args, **kwargs)

    def get_bytes_io_file(self):
        file = BytesIO()
        file.write(self._get_contents())
        file.seek(0)
        return file

    def __len__(self):
        return sum(len(s[0]) + len(s[1]) + len(s[2]) for s in self._content)

    def _get_contents(self):
        contents = b''
        for segment in self._content:
            contents += segment[0]
            contents += segment[1]
            contents += segment[2]
        return contents

    def _get_index_contents(self):
        contents = b''
        for segment in self._content:
            lead_in = segment[0]
            if len(lead_in) >= 4:
                lead_in = b'TDSh' + lead_in[4:]
            contents += lead_in
            contents += segment[1]
        return contents


class BytesIoTestFile(GeneratedFile):
    def load(self, *args, **kwargs):
        file = self.get_bytes_io_file()
        return tdms.TdmsFile(file, *args, **kwargs)


def _hex_to_bytes(hex_data):
    """ Converts a string of hex to a byte array
    """
    return binascii.unhexlify(
        hex_data.replace(" ", "").replace("\n", "").encode('utf-8'))


def compare_arrays(actual_data, expected_data):
    try:
        np.testing.assert_almost_equal(actual_data, expected_data)
    except TypeError:
        # Cannot compare given types
        assert len(actual_data) == len(expected_data)
        for (actual, expected) in zip(actual_data, expected_data):
            assert actual == expected



"""Test type reading and writing"""

from datetime import date, datetime
import io
import numpy as np
import pytest

from nptdms import types


@pytest.mark.parametrize(
    "time_string",
    [
        pytest.param('2019-11-08T18:47:00', id="standard timestamp"),
        pytest.param('0000-01-01T05:00:00', id="timestamp before TDMS epoch"),
        pytest.param('2019-11-08T18:47:00.123456', id="timestamp with microseconds"),
        pytest.param('1903-12-31T23:59:59.500', id="timestamp before TDMS epoch with microseconds"),
    ]
)
def test_timestamp_round_trip(time_string):
    expected_datetime = np.datetime64(time_string)

    timestamp = types.TimeStamp(expected_datetime)
    data_file = io.BytesIO(timestamp.bytes)

    read_datetime = types.TimeStamp.read(data_file).as_datetime64()

    assert expected_datetime == read_datetime


def test_timestamp_from_datetime():
    """Test timestamp from built in datetime value"""

    input_datetime = datetime(2019, 11, 8, 18, 47, 0)
    expected_datetime = np.datetime64('2019-11-08T18:47:00')

    timestamp = types.TimeStamp(input_datetime)
    data_file = io.BytesIO(timestamp.bytes)

    read_datetime = types.TimeStamp.read(data_file)

    assert expected_datetime == read_datetime.as_datetime64()


def test_timestamp_from_date():
    """Test timestamp from built in date value"""

    input_datetime = date(2019, 11, 8)
    expected_datetime = np.datetime64('2019-11-08T00:00:00')

    timestamp = types.TimeStamp(input_datetime)
    data_file = io.BytesIO(timestamp.bytes)

    read_datetime = types.TimeStamp.read(data_file)

    assert expected_datetime == read_datetime.as_datetime64()



import numpy as np
import pytest
from hypothesis import (given, strategies, settings)

from nptdms import thermocouples
import thermocouples_reference


settings.register_profile("thermocouples", deadline=None, max_examples=100)
settings.load_profile("thermocouples")


reference_thermocouple_b = thermocouples_reference.thermocouples['B']
reference_thermocouple_e = thermocouples_reference.thermocouples['E']
reference_thermocouple_j = thermocouples_reference.thermocouples['J']
reference_thermocouple_k = thermocouples_reference.thermocouples['K']
reference_thermocouple_n = thermocouples_reference.thermocouples['N']
reference_thermocouple_r = thermocouples_reference.thermocouples['R']
reference_thermocouple_s = thermocouples_reference.thermocouples['S']
reference_thermocouple_t = thermocouples_reference.thermocouples['T']


def test_scale_temperature_to_voltage():
    thermocouple = thermocouples.Thermocouple(
        forward_polynomials=[
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(None, 10),
                coefficients=[0.0, 1.0]),
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(10, 20),
                coefficients=[1.0, 2.0]),
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(20, None),
                coefficients=[2.0, 3.0]),
        ],
        inverse_polynomials=[]
    )
    voltages = thermocouple.celsius_to_mv(np.array([0.0, 9.0, 10.0, 11.0, 19.0, 20.0, 21.0]))
    np.testing.assert_almost_equal(voltages, np.array([0.0, 9.0, 21.0, 23.0, 39.0, 62.0, 65.0]))


def test_scale_voltage_to_temperature():
    thermocouple = thermocouples.Thermocouple(
        forward_polynomials=[],
        inverse_polynomials=[
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(None, 10),
                coefficients=[0.0, 1.0]),
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(10, 20),
                coefficients=[1.0, 2.0]),
            thermocouples.Polynomial(
                applicable_range=thermocouples.Range(20, None),
                coefficients=[2.0, 3.0]),
        ]
    )
    temperatures = thermocouple.mv_to_celsius(np.array([0.0, 9.0, 10.0, 11.0, 19.0, 20.0, 21.0]))
    np.testing.assert_almost_equal(temperatures, np.array([0.0, 9.0, 21.0, 23.0, 39.0, 62.0, 65.0]))


def test_non_contiguous_forward_polynomial_ranges():
    with pytest.raises(ValueError) as exc_info:
        _ = thermocouples.Thermocouple(
            forward_polynomials=[
                thermocouples.Polynomial(
                    applicable_range=thermocouples.Range(None, 10),
                    coefficients=[0.0, 1.0]),
                thermocouples.Polynomial(
                    applicable_range=thermocouples.Range(11, None),
                    coefficients=[0.0, 1.0]),
            ],
            inverse_polynomials=[]
        )
    assert "Polynomial ranges must be contiguous" in str(exc_info.value)


def test_non_contiguous_inverse_polynomial_ranges():
    with pytest.raises(ValueError) as exc_info:
        _ = thermocouples.Thermocouple(
            forward_polynomials=[],
            inverse_polynomials=[
                thermocouples.Polynomial(
                    applicable_range=thermocouples.Range(None, 10),
                    coefficients=[0.0, 1.0]),
                thermocouples.Polynomial(
                    applicable_range=thermocouples.Range(11, None),
                    coefficients=[0.0, 1.0]),
            ]
        )
    assert "Polynomial ranges must be contiguous" in str(exc_info.value)


def test_no_range():
    with pytest.raises(ValueError) as exc_info:
        _ = thermocouples.Range(None, None)
    assert "At least one of start and end must be provided" in str(exc_info.value)


@pytest.mark.parametrize("start,end", [(2.0, 1.0), (1.0, 1.0)])
def test_invalid_range(start, end):
    with pytest.raises(ValueError) as exc_info:
        _ = thermocouples.Range(start, end)
    assert "start must be less than end" in str(exc_info.value)


@given(temperature=strategies.floats(0.0, 1820.0))
def test_type_b_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_b, thermocouples.type_b, temperature)


@given(voltage=strategies.floats(0.291, 13.820))
def test_type_b_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_b, thermocouples.type_b, voltage, max_error=0.03)


@given(temperature=strategies.floats(-270.0, 1000.0))
def test_type_e_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_e, thermocouples.type_e, temperature)


@given(voltage=strategies.floats(-8.825, 76.372))
def test_type_e_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_e, thermocouples.type_e, voltage, max_error=0.03)


@given(temperature=strategies.floats(-210.0, 1200.0))
def test_type_j_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_j, thermocouples.type_j, temperature)


@given(voltage=strategies.floats(-8.095, 69.553))
def test_type_j_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_j, thermocouples.type_j, voltage, max_error=0.05)


@given(temperature=strategies.floats(-270.000, 1372.000))
def test_type_k_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_k, thermocouples.type_k, temperature)


@given(voltage=strategies.floats(-5.891, 54.886))
def test_type_k_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_k, thermocouples.type_k, voltage, max_error=0.06)


@given(temperature=strategies.floats(-270.000, 1300.0))
def test_type_n_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_n, thermocouples.type_n, temperature)


@given(voltage=strategies.floats(-3.990, 47.512))
def test_type_n_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_n, thermocouples.type_n, voltage, max_error=0.04)


@given(temperature=strategies.floats(-50.000, 1768.1))
def test_type_r_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_r, thermocouples.type_r, temperature)


@given(voltage=strategies.floats(-0.226, 21.102))
def test_type_r_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_r, thermocouples.type_r, voltage, max_error=0.02)


@given(temperature=strategies.floats(-50.000, 1768.1))
def test_type_s_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_s, thermocouples.type_s, temperature)


@given(voltage=strategies.floats(-0.235, 18.693))
def test_type_s_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_s, thermocouples.type_s, voltage, max_error=0.02)


@given(temperature=strategies.floats(-270.000, 400.000))
def test_type_t_temperature_to_voltage(temperature):
    _test_temperature_to_voltage(reference_thermocouple_t, thermocouples.type_t, temperature)


@given(voltage=strategies.floats(-5.603, 20.871))
def test_type_t_voltage_to_temperature(voltage):
    _test_voltage_to_temperature(reference_thermocouple_t, thermocouples.type_t, voltage, max_error=0.04)


def _test_temperature_to_voltage(reference_thermocouple, thermocouple, temperature):
    reference_voltage = reference_thermocouple.emf_mVC(temperature, Tref=0.0)
    voltage = thermocouple.celsius_to_mv(temperature)
    assert abs(voltage - reference_voltage) < 1.0E-6


def _test_voltage_to_temperature(reference_thermocouple, thermocouple, voltage, max_error):
    reference_temperature = reference_thermocouple.inverse_CmV(voltage, Tref=0.0)
    temperature = thermocouple.mv_to_celsius(voltage)
    assert abs(temperature - reference_temperature) < max_error + 1.0E-6



"""Test TdmsSegment"""

import os
import tempfile
from collections import OrderedDict
from datetime import datetime
from io import BytesIO

import numpy as np
import pytest
from nptdms.tdms import TdmsFile
from nptdms.types import *
from nptdms.writer import (ChannelObject, GroupObject, RootObject, TdmsSegment,
                           TdmsWriter, read_properties_dict)


class StubObject(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


def test_write_leadin_with_one_channel():
    data_type = StubObject(size=4)

    channel = StubObject(
        path="",
        has_data=True,
        data=[0] * 10,
        data_type=data_type)

    toc = ["kTocMetaData", "kTocRawData", "kTocNewObjList"]
    metadata_size = 12

    segment = TdmsSegment([channel])
    leadin = segment.leadin(toc, metadata_size)

    expected_values = [
        Bytes(b'TDSm'),
        Int32(14),  # TOC bitmask
        Int32(4712),  # TDMS version
        Uint64(52),  # Next segment offset
        Uint64(12),  # Raw data offset
        ]

    _assert_sequence_equal(leadin, expected_values)


def test_write_leadin_with_object_without_data():
    channel = StubObject(
        path="",
        has_data=False)

    toc = ["kTocMetaData", "kTocRawData", "kTocNewObjList"]
    metadata_size = 12

    segment = TdmsSegment([channel])
    leadin = segment.leadin(toc, metadata_size)

    expected_values = [
        Bytes(b'TDSm'),
        Int32(14),  # TOC bitmask
        Int32(4712),  # TDMS version
        Uint64(12),  # Next segment offset
        Uint64(12),  # Raw data offset
        ]

    _assert_sequence_equal(leadin, expected_values)


def test_write_metadata_with_properties():
    data_type = StubObject(enum_value=3)

    # Use an ordered dict for properties so that
    # the order of properties in metadata is guaranteed
    properties = OrderedDict()
    properties["prop1"] = String("foo")
    properties["prop2"] = Int32(42)

    channel = StubObject(
        path="channel_path",
        has_data=True,
        data=[1] * 10,
        data_type=data_type,
        properties=properties)

    segment = TdmsSegment([channel])
    metadata = segment.metadata()

    expected_values = [
        Uint32(1),  # Number of objects
        String("channel_path"),
        Uint32(20),  # Length of raw data index in bytes
        Int32(3),  # Data type
        Uint32(1),  # Array dimension
        Uint64(10),  # Number of values
        Uint32(2),  # Number of properties
        String("prop1"),  # Property name
        Int32(0x20),
        String("foo"),
        String("prop2"),
        Int32(3),
        Int32(42),
        ]

    _assert_sequence_equal(metadata, expected_values)


def test_write_metadata_with_no_data():
    obj = StubObject(
        path="object_path",
        has_data=False,
        properties={})

    segment = TdmsSegment([obj])
    metadata = segment.metadata()

    expected_values = [
        Uint32(1),  # Number of objects
        String("object_path"),
        Bytes(b'\xFF\xFF\xFF\xFF'),  # Raw data index
        Uint32(0),  # Number of properties
        ]

    _assert_sequence_equal(metadata, expected_values)


def test_properties_are_converted_to_tdms_types():
    test_time = datetime.utcnow()

    properties = {
        "prop1": Int32(1),
        "prop2": 2,
        "prop3": "foo",
        "prop4": True,
        "prop5": 3.142,
        "prop6": test_time,
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["prop1"] == Int32(1)
    assert tdms_properties["prop2"] == Int32(2)
    assert tdms_properties["prop3"] == String("foo")
    assert tdms_properties["prop4"] == Boolean(True)
    assert tdms_properties["prop5"] == DoubleFloat(3.142)
    assert tdms_properties["prop6"] == TimeStamp(test_time)


def test_datetime_converted_when_it_only_has_date_part():
    test_time = np.datetime64('2017-11-19')

    properties = {
        "time_prop": test_time,
    }

    tdms_properties = read_properties_dict(properties)

    assert tdms_properties["time_prop"] == TimeStamp(test_time)


def test_writing_long_integer_properties():
    properties = {
        "prop1": 2147483647,
        "prop2": 2147483648,
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["prop1"] == Int32(2147483647)
    assert tdms_properties["prop2"] == Int64(2147483648)


def test_writing_properties_with_numpy_typed_values():
    properties = {
        "int32prop": np.int32(32),
        "int64prop": np.int64(64),
        "float32prop": np.float32(32.0),
        "float64prop": np.float64(64.0),
        "boolprop": np.bool_(True),
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["int32prop"] == Int32(32)
    assert tdms_properties["int64prop"] == Int64(64)
    assert tdms_properties["float32prop"] == SingleFloat(32.0)
    assert tdms_properties["float64prop"] == DoubleFloat(64.0)
    assert tdms_properties["boolprop"] == Boolean(True)


def test_error_raised_when_cannot_convert_property_value():
    properties = {
        "prop1": None
    }

    with pytest.raises(TypeError):
        read_properties_dict(properties)


@pytest.mark.parametrize(
    "data",
    [
        np.array(0.0, dtype=np.dtype('float32')),
        np.zeros((2, 2), dtype=np.dtype('float32')),
        np.zeros((2, 2, 2), dtype=np.dtype('float32')),
    ]
)
def test_channel_with_invalid_data(data):
    with pytest.raises(ValueError) as exc_info:
        _ = ChannelObject("Group", "Channel", data, {})
    assert str(exc_info.value) == "Channel data must be a 1d array"


def _assert_sequence_equal(values, expected_values):
    position = 1
    expected_values = iter(expected_values)
    for val in values:
        try:
            expected = next(expected_values)
        except StopIteration:
            raise ValueError(
                "Expected end of sequence at position %d but found: %r" %
                (position, val))
        assert val == expected, "Expected %r to equal %r at position %d" % (val, expected, position)
        position += 1
    try:
        expected = next(expected_values)
        raise ValueError(
            "Expected %r at position %d but found end of sequence" %
            (expected, position))
    except StopIteration:
        pass


def test_defragment_files():
    buf = BytesIO()
    with TdmsWriter(buf) as file:
        file.write_segment([
            RootObject(properties={"file": "file1"}),
            GroupObject("group1"),
            ChannelObject("group1", "channel1", np.linspace(0, 1))
        ])
    buf.seek(0, os.SEEK_SET)

    target_buf = BytesIO()
    TdmsWriter.defragment(buf, target_buf)


def test_write_data_stream_with_index():
    data_file = BytesIO()
    index_file = BytesIO()
    with TdmsWriter(data_file, index_file=index_file) as file:
        file.write_segment([
            RootObject(properties={"file": "file1"}),
            GroupObject("group1"),
            ChannelObject("group1", "channel1", np.linspace(0, 1))
        ])

    data_file.seek(0, 0)
    index_file.seek(0, 0)
    assert len(data_file.read()) > 0
    assert len(index_file.read()) > 0


def test_write_data_stream_without_index():
    data_file = BytesIO()
    index_file = BytesIO()
    with TdmsWriter(data_file) as file:
        file.write_segment([
            RootObject(properties={"file": "file1"}),
            GroupObject("group1"),
            ChannelObject("group1", "channel1", np.linspace(0, 1))
        ])

    data_file.seek(0, 0)
    index_file.seek(0, 0)
    assert len(data_file.read()) > 0
    assert len(index_file.read()) == 0


def test_write_and_store_index_file():
    directory = tempfile.mkdtemp()
    tdms_path = os.path.join(directory, 'test_file.tdms')

    writer = TdmsWriter(tdms_path, index_file=True)
    with writer as file:
        file.write_segment([
            RootObject(properties={"file": "file1"}),
            GroupObject("group1"),
            ChannelObject("group1", "channel1", np.linspace(0, 1))
        ])

    assert os.path.isfile(tdms_path + "_index")
